# -*- coding: utf-8 -*-

"""
StochANNPy (STOCHAstic Artificial Neural Network for PYthon) provides
user-friendly routines compatible with Scikit-Learn for stochastic learning.

Author: Keurfon Luu <keurfon.luu@mines-paristech.fr>
License: MIT
"""

from .bayesian_neural_network import BNNClassifier
from .evolutionary_neural_network import ENNClassifier


__all__ = [ "BNNClassifier", "ENNClassifier" ]
__version__ = "0.0.1b2"


# -*- coding: utf-8 -*-

"""
Author: Keurfon Luu <keurfon.luu@mines-paristech.fr>
License: MIT
"""

import numpy as np
from abc import ABCMeta, abstractmethod
from six import with_metaclass
from sklearn.base import BaseEstimator
from sklearn.utils import check_X_y, check_array, column_or_1d
from sklearn.utils.validation import check_is_fitted
from sklearn.preprocessing import LabelBinarizer

__all__ = [ "BaseNeuralNetwork" ]


class BaseNeuralNetwork(with_metaclass(ABCMeta, BaseEstimator)):
    """
    Base class for artifical neural network.
    
    Do not use this class, please use derived classes.
    
    Parameters
    ----------
    hidden_layer_sizes : tuple or list, length = n_layers-2, default (10,)
        The ith element represents the number of neurons in the ith hidden
        layer.
    activation : {'logistic', 'tanh', 'relu'}, default 'relu'
        Activation function the hidden layer.
        - 'logistic', the logistic sigmoid function.
        - 'tanh', the hyperbolic tan function.
        - 'relu', the REctified Linear Unit function.
    alpha : scalar, optional, default 0.
        L2 penalty (regularization term) parameter.
    max_iter : int, optional, default 100
        Maximum number of iterations.
    bounds : scalar, optional, default 1.
        Search space boundaries for initialization.
    random_state : int, optional, default None
        Seed for random number generator.
    """
    
    @abstractmethod
    def __init__(self, hidden_layer_sizes = (10,), activation = "tanh", alpha = 0.,
                 max_iter = 100, bounds = 1., random_state = None):
        self.hidden_layer_sizes = hidden_layer_sizes
        self.activation = activation
        self.alpha = alpha
        self.max_iter = max_iter
        self.bounds = bounds
        self.random_state = random_state
            
    def _validate_base_hyperparameters(self):
        if isinstance(self.hidden_layer_sizes, tuple) or isinstance(self.hidden_layer_sizes, list):
            if not np.all([ isinstance(l, int) and l > 0 for l in self.hidden_layer_sizes ]):
                raise ValueError("hidden_layer_sizes must contains positive integers only, got %s" % self.hidden_layer_sizes)
        else:
            raise ValueError("hidden_layer_sizes must be a tuple or a list")
        if not isinstance(self.activation, str) and self.activation not in [ "logistic", "tanh" ]:
            raise ValueError("activation must either be 'logistic' or 'tanh', got %s" % self.activation)
        if not isinstance(self.alpha, float) and not isinstance(self.alpha, int) or self.alpha < 0.:
            raise ValueError("alpha must be positive, got %s" % self.alpha)
        if not isinstance(self.max_iter, int) or self.max_iter <= 0:
            raise ValueError("max_iter must be a positive integer, got %s" % self.max_iter)
        if not isinstance(self.bounds, float) and not isinstance(self.bounds, int) or self.bounds <= 0.:
            raise ValueError("bounds must be positive, got %s" % self.bounds)
        if self.random_state is not None:
            np.random.seed(self.random_state)

    def _validate_input(self, X, y):
        X, y = check_X_y(X, y)
        if y.ndim == 2 and y.shape[1] == 1:
            y = column_or_1d(y, warn = True)
            
        self._label_binarizer = LabelBinarizer()
        self._label_binarizer.fit(y)
        self.classes_ = self._label_binarizer.classes_
        y = self._label_binarizer.transform(y)
        return X, y
    
    def _initialize(self, X, y):
        # Make sure self.hidden_layer_sizes is a list
        hidden_layer_sizes = self.hidden_layer_sizes
        if not hasattr(hidden_layer_sizes, "__iter__"):
            hidden_layer_sizes = [ hidden_layer_sizes ]
        hidden_layer_sizes = list(hidden_layer_sizes)
        
        # Check that X and y have correct shape
        X, y = self._validate_input(X, y)
        self.ymat_ = np.array(y)
        self.n_samples_, self.n_features_ = X.shape
        self.n_outputs_ = y.shape[1]
        self.n_layers_ = len(hidden_layer_sizes) + 2
        self.layer_units_ = ( [ self.n_features_ ] + hidden_layer_sizes + [ self.n_outputs_ ] )
        
        # Store meta information for the parameters
        self.coef_indptr_ = []
        start = 0
        for i in range(self.n_layers_-1):
            n_fan_in, n_fan_out = self.layer_units_[i]+1, self.layer_units_[i+1]
            end = start + (n_fan_in * n_fan_out)
            self.coef_indptr_.append((start, end, (n_fan_out, n_fan_in)))
            start = end
            
        # Initialize functions
        self._init_functions()
        return X, y
    
    def _predict(self, X):
        """
        Predict using the trained model.
        
        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Input data.
        
        Returns
        -------
        h : ndarray
            Output layer activation values.
        """
        check_is_fitted(self, "coefs_")
        X = check_array(X)
        unpacked_coefs = self.coefs_
        Z, activations = self._forward_pass(unpacked_coefs, X)
        return activations[-1]
    
    def _forward_pass(self, unpacked_coefs, X):
        """
        Perform a feedforward pass on the network.
        
        Parameters
        ----------
        unpacked_coefs : list of ndarray
            The ith element of the list holds the biases and weights of the
            ith layer.
        X : ndarray of shape (n_samples, n_features)
            Input data.
            
        Returns
        -------
        Z : list of ndarray
            Layers propagation matrices.
        activations : list of ndarray
            Layers activation values.
        """
        Z = [ np.array([]) ]
        activations = [ np.hstack((np.ones((X.shape[0],1)), X)) ]
        for i in range(1, self.n_layers_-1):
            Z.append(np.dot(activations[i-1], unpacked_coefs[i-1].transpose()))
            activations.append(np.hstack((np.ones((Z[i].shape[0],1)), self._func(Z[i]))))
        Z.append(np.dot(activations[-1], unpacked_coefs[-1].transpose()))
        activations.append(self._output_func(Z[-1]))
        activations[-1] = np.clip(activations[-1], 1e-10, 1 - 1e-10)
        return Z, activations
    
    def _backprop(self, Z, activations, unpacked_coefs):
        """
        Perform the backpropagation algorithm to compute the gradient.
        
        Parameters
        ----------
        Z : list of ndarray
            Layers propagation matrices.
        activations : list of ndarray
            Layers activation values.
        unpacked_coefs : list of ndarray
            The ith element of the list holds the biases and weights of the
            ith layer.
        
        Returns
        -------
        grad : list of ndarray
            The ith element of the list holds the gradient of the biases and
            weights of the ith layer.
        """
        sigma = [ np.array([]) for i in range(self.n_layers_) ]
        sigma[-1] = activations[-1] - self.ymat_
        for i in range(self.n_layers_-2, 0, -1):
            sigma[i] = np.dot(sigma[i+1], unpacked_coefs[i]) * self._fprime(np.hstack((np.ones((Z[i].shape[0],1)), Z[i])))
            sigma[i] = sigma[i][:,1:]
        
        # Accumulate gradient
        delta = []
        for i in range(self.n_layers_-1):
            delta.append(np.dot(sigma[i+1].T, activations[i]))
            
        # Unroll gradient
        grad = []
        for i in range(self.n_layers_-1):
            grad.append(delta[i] / self.n_samples_ + self.alpha / self.n_samples_ * np.hstack((np.zeros((unpacked_coefs[i].shape[0], 1)), unpacked_coefs[i][:,1:])))        
        return grad
    
    def _loss(self, packed_coefs, X):
        """
        Compute the log loss function.
        
        Parameters
        ----------
        packed_coefs : ndarray
            Neural network biases and weights.
        X : ndarray of shape (n_samples, n_features)
            Input data.
            
        Returns
        -------
        loss : scalar
            Loss function value.
        """
        unpacked_coefs = self._unpack(packed_coefs)                 # Unpack
        Z, activations = self._forward_pass(unpacked_coefs, X)      # Forward pass
        values = np.sum(np.array([ np.dot(s[:,1:].ravel(), s[:,1:].ravel()) \
                                  for s in unpacked_coefs ]))       # L2 regularization penalty
        loss = np.sum(-self.ymat_*np.log(activations[-1]) \
                      - (1.-self.ymat_)*np.log(1.-activations[-1])) # Log loss function
        loss += 0.5 * self.alpha * values
        loss /= self.n_samples_
        return loss
    
    def _grad(self, packed_coefs, X):
        """
        Compute the gradient of the loss function.
        
        Parameters
        ----------
        packed_coefs : ndarray
            Neural network biases and weights.
        X : ndarray of shape (n_samples, n_features)
            Input data.
            
        Returns
        -------
        grad : ndarray
            Gradient of neural network biases and weights.
        """
        unpacked_coefs = self._unpack(packed_coefs)                 # Unpack
        Z, activations = self._forward_pass(unpacked_coefs, X)      # Forward pass
        grad = self._backprop(Z, activations, unpacked_coefs)       # Backward propagation
        grad = self._pack(grad)                                     # Pack gradient coefficients
        return grad
    
    def _init_functions(self):
        if self.activation == "logistic":
            self._func = self._sigmoid
            self._fprime = self._sigmoid_grad
        elif self.activation == "tanh":
            self._func = self._tanh
            self._fprime = self._tanh_grad
        elif self.activation == "relu":
            self._func = self._relu
            self._fprime = self._relu_grad
        if self._label_binarizer.y_type_ == "multiclass":
            self._output_func = self._softmax
        else:
            self._output_func = self._sigmoid
    
    def _init_coefs(self):
        coefs_init = []
        for i in range(self.n_layers_-1):
            if self.activation == "logistic":
                init_bound = np.sqrt(2. / (self.layer_units_[i] + self.layer_units_[i+1] + 1))
            else:
                init_bound = np.sqrt(6. / (self.layer_units_[i] + self.layer_units_[i+1] + 1))
            coefs_init.append(self._rand(self.layer_units_[i+1], self.layer_units_[i]+1, init_bound))
        return coefs_init
    
    def _rand(self, n1, n2, init_bound = 1.):
        """
        Generate uniform array between -init_bound and init_bound.
        
        Parameters
        ----------
        n1 : int
            First dimension of array.
        n2 : int
            Second dimension of array.
        init_bound : float, optional, default 1.
            Random number absolute maximum.
        
        Returns
        -------
        r : ndarray of shape (n1, n2)
            Uniform array between -init_bound and init_bound.
        """
        rnd = 2. * np.random.rand(n1, n2) - 1.
        return init_bound * rnd
    
    def _sigmoid(self, x):
        out = np.array(x, dtype = np.float64)
        return 0.5 * (1. + np.tanh(0.5*out))
        
    def _sigmoid_grad(self, x):
        sig = self._sigmoid(x)
        return sig * (1 - sig)
    
    def _tanh(self, x):
        out = np.array(x, dtype = np.float64)
        return np.tanh(out)
    
    def _tanh_grad(self, x):
        return 1. - self._tanh(x)**2
    
    def _relu(self, x):
        return np.clip(x, 0, np.finfo(x.dtype).max)
    
    def _relu_grad(self, x):
        return 1. * (np.array(x) > 0.)
    
    def _softmax(self, x):
        tmp = x - x.max(axis = 1)[:,np.newaxis]
        x = np.exp(tmp)
        x /= x.sum(axis = 1)[:,np.newaxis]
        return x
    
    def _pack(self, unpacked_parameters):
        """
        Pack the coefficients into a 1-D array.
        
        Parameters
        ----------
        unpacked_parameters : list of ndarray
            The ith element of the list holds the biases and weights of the
            ith layer.
        
        Returns
        -------
        packed_parameters : ndarray
            Neural network biases and weights.
        """
        return np.hstack([ l.ravel(order = "F") for l in unpacked_parameters ])

    def _unpack(self, packed_parameters):
        """
        Unpack the coefficients into a list of NumPy arrays.
        
        Parameters
        ----------
        packed_parameters : ndarray
            Neural network biases and weights.
        
        Returns
        -------
        unpacked_parameters : list of ndarray
            The ith element of the list holds the biases and weights of the
            ith layer.
        """
        unpacked_parameters = []
        for i in range(self.n_layers_ - 1):
            start, end, shape = self.coef_indptr_[i]
            unpacked_parameters.append(np.reshape(packed_parameters[start:end], shape, order = "F"))
        return unpacked_parameters


# -*- coding: utf-8 -*-

"""
Author: Keurfon Luu <keurfon.luu@mines-paristech.fr>
License: MIT
"""

import numpy as np
from sklearn.base import ClassifierMixin
from ._base_neural_network import BaseNeuralNetwork
from stochopy import Evolutionary

__all__ = [ "ENNClassifier" ]


class ENNClassifier(BaseNeuralNetwork, ClassifierMixin):
    """
    Evolutionary neural network classifier.
    
    This model optimizes the log-loss function using Differential Evolution,
    Particle Swarm Optimization of Covariance Matrix Adaptation - Evolution
    Strategy.
    
    Parameters
    ----------
    hidden_layer_sizes : tuple or list, length = n_layers-2, default (10,)
        The ith element represents the number of neurons in the ith hidden
        layer.
    activation : {'logistic', 'tanh', 'relu'}, default 'relu'
        Activation function the hidden layer.
        - 'logistic', the logistic sigmoid function.
        - 'tanh', the hyperbolic tan function.
        - 'relu', the REctified Linear Unit function.
    alpha : scalar, optional, default 0.
        L2 penalty (regularization term) parameter.
    max_iter : int, optional, default 100
        Maximum number of iterations.
    solver : {'de', 'pso', 'cmaes'}, default 'pso'
        Evolutionary Algorithm optimizer.
    popsize : int, optional, default 10
        Population size.
    w : scalar, optional, default 0.72
        Inertial weight. Only used when solver = 'pso'.
    c1 : scalar, optional, default 1.49
        Cognition parameter. Only used when solver = 'pso'.
    c2 : scalar, optional, default 1.49
        Sociability parameter. Only used when solver = 'pso'.
    l : scalar, optional, default 0.1
        Velocity clamping percentage. Only used when solver = 'pso'.
    F : scalar, optional, default 1.
        Differential weight. Only used when solver = 'de'.
    CR : scalar, optional, default 0.5
        Crossover probability. Only used when solver = 'de'.
    sigma : scalar, optional, default 1.
        Step size. Only used when solver = 'cmaes'.
    mu_perc : scalar, optional, default 0.5
        Number of parents as a percentage of population size. Only used
        when solver = 'cmaes'.
    eps1 : scalar, optional, default 1e-8
        Minimum change in best individual.
    eps2 : scalar, optional, default 1e-8
        Minimum objective function precision.
    bounds : scalar, optional, default 1.
        Search space boundaries for initialization.
    random_state : int, optional, default None
        Seed for random number generator.
    
    Examples
    --------
    Import the module and initialize the classifier:
    
    >>> import numpy as np
    >>> from stochannpy import ENNClassifier
    >>> clf = ENNClassifier(hidden_layer_sizes = (5,))
    
    Fit the training set:
    
    >>> clf.fit(X_train, y_train)
    
    Predict the test set:
    
    >>> ypred = clf.predict(X_test)
    
    Compute the accuracy:
    
    >>> print(np.mean(ypred == y_test))
    """

    def __init__(self, hidden_layer_sizes = (10,), max_iter = 100, alpha = 0.,
                 activation = "relu", solver = "pso", popsize = 10,
                 w = 0.72, c1 = 1.49, c2 = 1.49, l = 0.1, F = 1., CR = 0.5,
                 sigma = 1., mu_perc = 0.5, eps1 = 1e-8, eps2 = 1e-8,
                 bounds = 1., random_state = None):
        super(ENNClassifier, self).__init__(
            hidden_layer_sizes = hidden_layer_sizes,
            activation = activation,
            alpha = alpha,
            max_iter = max_iter,
            bounds = bounds,
            random_state = random_state,
        )
        self.solver = solver
        self.popsize = int(popsize)
        self.eps1 = eps1
        self.eps2 = eps2
        self.w = w
        self.c1 = c1
        self.c2 = c2
        self.l = l
        self.F = F
        self.CR = CR
        self.sigma = sigma
        self.mu_perc = mu_perc
        
    def _validate_hyperparameters(self):
        self._validate_base_hyperparameters()
        if not isinstance(self.solver, str) or self.solver not in [ "pso", "de", "cmaes" ]:
            raise ValueError("solver must either be 'pso', 'de' or 'cmaes', got %s" % self.solver)
        if not isinstance(self.popsize, int) or self.popsize < 2:
            raise ValueError("popsize must be an integer > 1, got %s" % self.popsize)
        if not isinstance(self.eps1, float) and not isinstance(self.eps1, int) or self.eps1 < 0.:
            raise ValueError("eps1 must be positive, got %s" % self.eps1)
        if not isinstance(self.eps2, float) and not isinstance(self.eps2, int):
            raise ValueError("eps2 must be an integer or float, got %s" % self.eps2)
    
    def fit(self, X, y):
        """
        Fit the model to data matrix X and target y.
        
        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Input data.
        y : ndarray of length n_samples
            Target values.
        """
        # Check inputs and initialize
        self._validate_hyperparameters()
        X, y = self._initialize(X, y)
        
        # Initialize boundaries
        n_dim = np.sum([ np.prod(i[-1]) for i in self.coef_indptr_ ])
        lower = np.full(n_dim, -self.bounds)
        upper = np.full(n_dim, self.bounds)
        
        # Optimize using PSO, DE or CMA-ES
        ea = Evolutionary(self._loss,
                          lower = lower,
                          upper = upper,
                          max_iter = self.max_iter,
                          popsize = self.popsize,
                          eps1 = self.eps1,
                          eps2 = self.eps2,
                          args = (X,))
        if self.solver == "de":
            packed_coefs, self.loss_ = ea.optimize(solver = "de",
                                                   F = self.F,
                                                   CR = self.CR)
        elif self.solver == "pso":
            packed_coefs, self.loss_ = ea.optimize(solver = "pso",
                                                   w = self.w,
                                                   c1 = self.c1,
                                                   c2 = self.c2,
                                                   l = self.l)
        elif self.solver == "cmaes":
            packed_coefs, self.loss_ = ea.optimize(solver = "cmaes",
                                                   sigma = self.sigma,
                                                   mu_perc = self.mu_perc)
        self.coefs_ = self._unpack(packed_coefs)
        self.flag_ = ea.flag
        self.n_iter_ = ea.n_iter
        self.n_eval_ = ea.n_eval        
        return self
    
    def predict(self, X):
        """
        Predict using the trained model.
        
        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Input data.
        
        Returns
        -------
        ypred : ndarray of length n_samples
            Predicted labels.
        """
        y_pred = self._predict(X)
        if self.n_outputs_ == 1:
            y_pred = y_pred.ravel()
        return self._label_binarizer.inverse_transform(y_pred)
    
    def predict_log_proba(self, X):
        """
        Log of probability estimates.
        
        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Input data.
        
        Returns
        -------
        yprob : ndarray of shape (n_samples, n_outputs)
            The ith row and jth column holds the log-probability of the ith
            sample to the jth class
        """
        return np.log(self.predict_proba(X))
    
    def predict_proba(self, X):
        """
        Probability estimates.
        
        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Input data.
        
        Returns
        -------
        yprob : ndarray of shape (n_samples, n_outputs)
            The ith row and jth column holds the probability of the ith sample
            to the jth class
        """
        y_pred = self._predict(X)
        if self.n_outputs_ == 1:
            y_pred = y_pred.ravel()
        if y_pred.ndim == 1:
            return np.vstack([1. - y_pred, y_pred]).transpose()
        else:
            return y_pred
    
    def score(self, X, y):
        """
        Compute accuracy score.
        
        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Input data.
        y : ndarray of length n_samples
            Target values.
            
        Returns
        -------
        acc : scalar
            Accuracy of prediction.
        """
        return np.mean(self.predict(X) == y)
    
    @property
    def weights(self):
        """
        list of ndarray
        Neural network weights. The ith element holds the weighs for the
        layer i.
        """
        return [ np.array(coefs[:,1:].T) for coefs in self.coefs_ ]
    
    @property
    def biases(self):
        """
        list of ndarray
        Neural network biases. The ith element holds the biases for the
        layer i.
        """
        return [ np.array(coefs[:,0]) for coefs in self.coefs_ ]
    
    @property
    def flag(self):
        """
        int
        Stopping criterion:
            - 0, best individual position changes less than eps1.
            - 1, maximum number of iterations is reached.
            - 2, NoEffectAxis (only when solver = 'cmaes').
            - 3, NoEffectCoord (only when solver = 'cmaes').
            - 4, ConditionCov (only when solver = 'cmaes').
            - 5, EqualFunValues (only when solver = 'cmaes').
            - 6, TolXUp (only when solver = 'cmaes').
            - 7, TolFun (only when solver = 'cmaes').
            - 8, TolX (only when solver = 'cmaes').
        """
        return self.flag_
    
    @property
    def n_iter(self):
        """
        int
        Number of iterations required to reach stopping criterion.
        """
        return self.n_iter_
    
    @property
    def n_eval(self):
        """
        int
        Number of function evaluations performed.
        """
        return self.n_eval_


# -*- coding: utf-8 -*-
 
from setuptools import setup, find_packages
import stochannpy

DISTNAME = "stochannpy"
DESCRIPTION = "StochANNPy"
LONG_DESCRIPTION = """StochANNPy (STOCHAstic Artificial Neural Network for PYthon) provides user-friendly routines compatible with Scikit-Learn for stochastic learning."""
AUTHOR = "Keurfon LUU"
AUTHOR_EMAIL = "keurfon.luu@mines-paristech.fr"
URL = "https://github.com/keurfonluu/stochannpy"
LICENSE = "MIT License"
REQUIREMENTS = [
    "numpy >= 1.9.0",
    "stochopy >= 1.1.2",
    "scikit-learn",
    "scipy",
    "matplotlib",
]
CLASSIFIERS = [
    "Programming Language :: Python",
    "Development Status :: 4 - Beta",
    "License :: OSI Approved :: MIT License",
    "Natural Language :: English",
    "Operating System :: OS Independent",
    "Intended Audience :: Science/Research",
    "Topic :: Scientific/Engineering",
    "Topic :: Scientific/Engineering :: Mathematics",
    "Topic :: Scientific/Engineering :: Artificial Intelligence",
    "Topic :: Scientific/Engineering :: Image Recognition",
]
 
if __name__ == "__main__":
    setup(
        name = DISTNAME,
        description = DESCRIPTION,
        long_description = LONG_DESCRIPTION,
        author = AUTHOR,
        author_email = AUTHOR_EMAIL,
        url = URL,
        license = LICENSE,
        install_requires = REQUIREMENTS,
        classifiers = CLASSIFIERS,
        version = stochannpy.__version__,
        packages = find_packages(),
        include_package_data = True,
    )


# -*- coding: utf-8 -*-

"""
Author: Keurfon Luu <keurfon.luu@mines-paristech.fr>
License: MIT
"""

import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import gaussian_kde
from sklearn.base import ClassifierMixin
from sklearn.utils.validation import check_array, check_is_fitted
from ._base_neural_network import BaseNeuralNetwork
from stochopy import MonteCarlo

__all__ = [ "BNNClassifier" ]


class BNNClassifier(BaseNeuralNetwork, ClassifierMixin):
    """
    Bayesian neural network classifier.
    
    This model samples the weights of the log-loss function using the
    Metropolis-Hastings algorithm or Hamiltonian (Hybrid) Monte-Carlo.
    
    Parameters
    ----------
    hidden_layer_sizes : tuple or list, length = n_layers-2, default (10,)
        The ith element represents the number of neurons in the ith hidden
        layer.
    activation : {'logistic', 'tanh', 'relu'}, default 'relu'
        Activation function the hidden layer.
        - 'logistic', the logistic sigmoid function.
        - 'tanh', the hyperbolic tan function.
        - 'relu', the REctified Linear Unit function.
    alpha : scalar, optional, default 0.
        L2 penalty (regularization term) parameter.
    max_iter : int, optional, default 1000
        Maximum number of iterations.
    sampler : {'mcmc', 'hastings', 'hmc', 'hamiltonian'}, default 'mcmc'
        Sampling method.
    stepsize : scalar, optional, default 1e-1
        If sampler = 'mcmc', standard deviation of gaussian perturbation.
        If sampler = 'hmc', leap-frog step size.
    n_leap : int, optional, default 10
        Number of leap-frog steps. Only used when sampler = 'hmc'.
    bounds : scalar, optional, default 1.
        Search space boundaries for initialization.
    random_state : int, optional, default None
        Seed for random number generator.
    
    Examples
    --------
    Import the module and initialize the classifier:
    
    >>> import numpy as np
    >>> from stochannpy import BNNClassifier
    >>> clf = BNNClassifier(hidden_layer_sizes = (5,))
    
    Fit the training set:
    
    >>> clf.fit(X_train, y_train)
    
    Predict the test set:
    
    >>> ypred = clf.predict(X_test)
    
    Compute the accuracy:
    
    >>> print(np.mean(ypred == y_test))
    """

    def __init__(self, hidden_layer_sizes = (10,), activation = "relu", alpha = 0.,
                 max_iter = 1000, sampler = "mcmc", stepsize = 1e-1, n_leap = 10,
                 bounds = 1., random_state = None):
        super(BNNClassifier, self).__init__(
            hidden_layer_sizes = hidden_layer_sizes,
            activation = activation,
            alpha = alpha,
            max_iter = max_iter,
            bounds = bounds,
            random_state = random_state,
        )
        self.sampler = sampler
        self.stepsize = stepsize
        self.n_leap = n_leap
            
    def _validate_hyperparameters(self):
        self._validate_base_hyperparameters()
        if not isinstance(self.sampler, str) or self.sampler not in [ "mcmc", "hastings", "hmc", "hamiltonian" ]:
            raise ValueError("sampler must either be 'mcmc', 'hastings', 'hmc' or 'hamiltonian', got %s" % self.sampler)
        if not isinstance(self.stepsize, float) and not isinstance(self.stepsize, int) or self.stepsize <= 0.:
            raise ValueError("stepsize must be positive, got %s" % self.stepsize)
        if not isinstance(self.n_leap, int) or self.n_leap <= 0:
            raise ValueError("n_leap must be a positive integer, got %s" % self.n_leap)
    
    def fit(self, X, y):
        """
        Fit the model to data matrix X and target y.
        
        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Input data.
        y : ndarray of length n_samples
            Target values.
        """
        # Check inputs and initialize
        self._validate_hyperparameters()
        X, y = self._initialize(X, y)
        
        # Initialize boundaries
        n_dim = np.sum([ np.prod(i[-1]) for i in self.coef_indptr_ ])
        lower = np.full(n_dim, -self.bounds)
        upper = np.full(n_dim, self.bounds)
        
        # Optimize using Hamiltonian Monte-Carlo
        mc = MonteCarlo(self._loss,
                        lower = lower,
                        upper = upper,
                        max_iter = self.max_iter,
                        args = (X,))
        if self.sampler in [ "hmc", "hamiltonian" ]:
            mc.sample(sampler = "hamiltonian",
                      fprime = self._grad,
                      stepsize = self.stepsize,
                      n_leap = self.n_leap,
                      args = (X,))
        elif self.sampler in [ "mcmc", "hastings" ]:
            mc.sample(sampler = "hastings",
                      stepsize = self.stepsize)
        self.models_ = mc.models
        self.energy_ = mc.energy
        self.acceptance_ratio_ = mc.acceptance_ratio
        self.n_iter_ = self.max_iter
        return self
    
    def predict(self, X, n_burnin = 0):
        """
        Predict using the trained model.
        
        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Input data.
        n_burnin : int, optional, default 0
            Number of samples corresponding to burn-in period (ignored in
            prediction).
        
        Returns
        -------
        ypred : ndarray of length n_samples
            Predicted labels.
        """
        y_pred = self._predict(X)
        y_pred = np.sum(self._integrate(y_pred, n_burnin), 0)
        if self.n_outputs_ == 1:
            y_pred = y_pred.ravel()
        return self._label_binarizer.inverse_transform(y_pred)
    
    def predict_log_proba(self, X, n_burnin = 0):
        """
        Log of probability estimates.
        
        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Input data.
        n_burnin : int, optional, default 0
            Number of samples corresponding to burn-in period (ignored in
            prediction).
        
        Returns
        -------
        yprob : ndarray of shape (n_samples, n_outputs)
            The ith row and jth column holds the log-probability of the ith
            sample to the jth class
        """
        return np.log(self.predict_proba(X, n_burnin))
    
    def predict_proba(self, X, n_burnin = 0):
        """
        Probability estimates.
        
        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Input data.
        n_burnin : int, optional, default 0
            Number of samples corresponding to burn-in period (ignored in
            prediction).
        
        Returns
        -------
        yprob : ndarray of shape (n_samples, n_outputs)
            The ith row and jth column holds the probability of the ith sample
            to the jth class
        """
        y_pred = self._predict(X)
        y_pred = np.sum(self._integrate(y_pred, n_burnin), 0)
        if self.n_outputs_ == 1:
            y_pred = y_pred.ravel()
        if y_pred.ndim == 1:
            return np.vstack([1. - y_pred, y_pred]).transpose()
        else:
            return y_pred
    
    def _predict(self, X):
        check_is_fitted(self, "models_")
        X = check_array(X)
        y_pred = [ np.array([]) for i in range(self.max_iter) ]
        for i in range(self.max_iter):
            unpacked_coefs = self._unpack(self.models_[:,i])
            Z, activations = self._forward_pass(unpacked_coefs, X)
            y_pred[i] = activations[-1]
        return y_pred
            
    def _integrate(self, y_pred, n_burnin = 0):
        weights = np.exp(-self.energy_[n_burnin:])
        weights /= np.sum(weights)
        y_pred = [ y_pred[n_burnin+i]*weights[i] for i in range(len(weights)) ]
        return y_pred
    
    def score(self, X, y):
        """
        Compute accuracy score.
        
        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Input data.
        y : ndarray of length n_samples
            Target values.
            
        Returns
        -------
        acc : scalar
            Accuracy of prediction.
        """
        return np.mean(self.predict(X) == y)
    
    def plot_coefs(self, n_burnin = 0, ignore_bias = True):
        """
        Plot the posterior distribution of the sampled coefficients (biases
        and weights).
        
        Parameters
        ----------
        n_burnin : int, optional, default 0
            Number of samples corresponding to burn-in period to ignore.
        ignore_bias : bool, optional, default True
            If False, plot the posterior distribution of the biases.
        """
        weights = self.weights
        if not ignore_bias:
            biases = self.biases
            
        for i in range(self.n_layers_-1):
            fig = plt.figure(figsize = (8, 8), facecolor = "white")
            ax1 = fig.add_subplot(1, 1, 1)
            
            for j in range(weights[i].shape[0]):
                wmin = np.min(weights[i][j,n_burnin:])
                wmax = np.max(weights[i][j,n_burnin:])
                aw = np.linspace(wmin, wmax, 101)
                kde = gaussian_kde(weights[i][j,n_burnin:])
                pdf = kde.pdf(aw)
                ax1.plot(aw, pdf, linestyle = "-")
            
            if not ignore_bias:
                for j in range(biases[i].shape[0]):
                    bmin = np.min(biases[i][j,n_burnin:])
                    bmax = np.max(biases[i][j,n_burnin:])
                    ab = np.linspace(bmin, bmax, 101)
                    kde = gaussian_kde(biases[i][j,n_burnin:])
                    pdf = kde.pdf(ab)
                    ax1.plot(ab, pdf, linestyle = ":")
    
    @property
    def weights(self):
        """
        list of ndarray
        Sampled neural network weights. The ith element holds all the weights
        sampled for the layer i.
        """
        weights = []
        for i in range(self.n_layers_-1):
            start, end, shape = self.coef_indptr_[i]
            nw = np.prod(shape)-shape[0]
            w_i = np.zeros((nw, self.max_iter))
            for j, k in enumerate(range(start+shape[0], end)):
                w_i[j,:] = self.models_[k,:]
            weights.append(np.array(w_i))
        return weights
    
    @property
    def biases(self):
        """
        list of ndarray
        Sampled neural network biases. The ith element holds all the biases
        sampled for the layer i.
        """
        biases = []
        for i in range(self.n_layers_-1):
            start, end, shape = self.coef_indptr_[i]
            b_i = np.zeros((shape[0], self.max_iter))
            for j, k in enumerate(range(start, start+shape[0])):
                b_i[j,:] = self.models_[k,:]
            biases.append(np.array(b_i))
        return biases
    
    @property
    def models(self):
        """
        ndarray of shape (n_dim, max_iter)
        Sampled models.
        """
        return self.models_
    
    @property
    def energy(self):
        """
        ndarray of shape (max_iter)
        Energy of sampled models.
        """
        return self.energy_
    
    @property
    def acceptance_ratio(self):
        """
        scalar between 0 and 1
        Acceptance ratio of sampler.
        """
        return self.acceptance_ratio_